#!/usr/bin/python
#
# BIND zone file parser
#
# Jason Vas Dias <jvdias@redhat.com> Sep 2004

import sys
import os
import commands
import datetime
import types
import string
if __name__ == "__main__":
    import __builtin__
    import gettext
    os.environ['SCB']='/usr/share/system-config-bind'
    os.putenv('SCB','/usr/share/system-config-bind')
    SCB='/usr/share/system-config-bind'
    gettext.bindtextdomain('bindconf', '/usr/share/locale')
    gettext.textdomain('bindconf')
    __builtin__.__dict__['_'] = gettext.gettext
    import Debug
    dbg = os.getenv('SCB_DEBUG')
    if dbg != None:
        Debug.DEBUG_CWD = SCB
        Debug.DEBUG = 1
        Debug.parseEnv(dbg,Debug)
import DNS
import RR

class Zone:
    origin=None
    filename=None
    contents=None
    records=[]
    files={}
    type=None
    clas=None
    ttl=None
    OUT=-1
    continuation=False
    errors=[]
    
    def parse(self):
        self.records=[]
        self.files={}
        token=''
        begin=0
        end=0        
        token=''
        tokens=[]
        token_offsets=[]
        inComment=False
        inQuote=False
        inMultiLine=False
        continuation=self.continuation
        p=None
        lines=0
        if( len(self.contents) <= 0):
            return
        if self.contents[-1] != '\n':
            self.contents += '\n'
        for c in self.contents:
            if (not inQuote) and (p != '\\') and (c == ';'):
                if len(token):
                    tokens.append(token)
                    token_offsets.append([begin,end])
                    token=''
                inComment=True
                end+=1
                p=';'
                continue
            if inComment:
                if c=='\n':
                    lines+=1
                    inComment=False
                    inQuote=False
                    p='\n'
                else:
                    end+=1
                    continue
            if  c == '"' and (p != '\\'):
                inQuote = not inQuote
            if (not inQuote) and (p != '\\') and (c=='('):
                inMultiLine=True
                if p in [' ', '\t'] and (len(token)==0) and (len(tokens)==0):
                    continuation=True
            if (not inQuote) and (p != '\\') and (c==')'):
                inMultiLine=False
            if (not inQuote) and (p != '\\') and (c in [' ','\t','\n','(',')']):
                if len(token):
                    tokens.append(token)
                    token_offsets.append([begin,end])
                    token=''                    
                if  (c == '\n'):
                    lines+=1
                    if (not inMultiLine) and len(tokens):
                        debug( 'RR: ',continuation,string.join(tokens,' '))
                        rr=RR.parse(self, continuation, tokens, token_offsets)
                        if( rr == None ):
                            self.errors.append( self.filename +':' + str(lines) +':\n'+_('WARNING : ignoring unrecognized record:\n%s.')% string.join(tokens,' '))
                        elif(rr == Zone.OUT):
                            self.errors.append( self.filename +':' + str(lines) +':\n'+_('WARNING : ignoring out-of-zone record:\n%s.') % string.join(tokens,' '))
                        elif rr.Type() == 'INCLUDE':
                            if( rr.origin != None ):
                                z_orig=rr.origin
                            else:
                                z_orig=self.origin
                            fn = rr.filename
                            if fn[0] != '/':
                                fn = self.nmdc.directory() + '/' + fn
                            z = Zone(fn, z_orig, self.nmdc, self.nmdcKey,  self.clas, self.cfgType, continuation)
                            for r in z.records:
                                if r.file == None:
                                    r.file=rr.filename
                                r.zone=self
                                if (r.name != None) and (not self.In(str(r.name))):
                                    self.errors.append( self.filename+ ':'+ str(lines) + ': ' + _('WARNING : ignoring INCLUDEd out-of-zone record:\n%s.') % string.join(r.tokens,' '))
                                else:
                                    r.index = len(self.records)
                                    self.records.append(r)                                    
                            self.files[fn]=z.contents
                            for f in z.files:
                                self.files[f]=z.files[f]
                            del z
                        tokens=[]
                        token_offsets=[]
                        continuation=False
                elif (p == '\n') and (c != '(') and (len(token)==0) and (len(tokens)==0):
                    continuation=True
            else:
                if len(token) == 0:
                    if ( p == "\n" ) and (c != '$'):
                        continuation=False
                    begin=end
                token += c
            p=c
            end+=1            
    
    def read(self):
        try:
            fd = os.open(self.filename, os.O_RDONLY)
            s = os.fstat(fd)
            self.contents = os.read(fd, s.st_size)        
            os.close(fd)
        except OSError, e :
            self.errors.append(e)
            return e
        self.parse()        
        return 0

    def load(self, not_include):
        self.contents=''
        self.records=[]
        self.files={}
        self.ttl=None
        self.errors=[]        
        if os.access(self.filename, os.R_OK) == 1:
            s=[0]
            if not_include and (self.cfgType != 'hint'): # named-checkzone can't handle root cache zones
                s=commands.getstatusoutput('/usr/sbin/named-checkzone ' + self.origin + ' ' + self.filename)
            if( s[0]!= 0) and len(s[1]):
                self.errors.append(s[1])
            else:
                self.read()
        elif self.cfgType != 'slave':
            self.errors.append(_('ERROR: Zone %s file %s does not exist.') %( self.zone_origin,self.filename ))
    
    def In(self,name):
        if name.find('*') > -1:
            na = DNS.split_dots(name)
            oa = DNS.split_dots(self.zone_origin)
            debug('In: name:',na, 'origin:',oa)
            ni = len(na)-1
            if ni < 0:
                ni = 0
            for i in range(len(oa)-1,-1,-1):
                if (na[ni] != '*') and (na[ni].lower() != oa[i].lower()):
                    return False
                if ni > 0:
                    ni -= 1
            return True
        else:
            return name[-(len(self.zone_origin)):].lower() == self.zone_origin.lower()

    def out(self,file=None):
        saved_contents=''
        if file != None:
            saved_contents = self.contents
            self.contents = self.files[file]            
        begin = 0
        end = 0
        if len(self.records) == 0:
            self.contents=''
            return self.contents
        i=0
        output=''
        last_index=-1
        last_name=None
        last_exists=len(self.records) and self.records[0].has_data()
        origin=self.zone_origin
        if ((self.contents == None) or (len(self.contents) <= 0)) and (self.ttl != None):
            # new zone. emit default TTL:
            output+='$TTL'+ ' ' + str(self.ttl) + "\n"
        for r in self.records:
            if r.file != file:
                 continue
            else:
                last_index = r.index
            if len(r.offsets):
                end=r.offsets[0][0]
                if last_exists :
                    if  not r.has_data():
                        while( (end > begin) and (self.contents[end] != '\n')):
                            end -= 1
                        if(self.contents[end]=='\n'):
                            end+=1
                    output +=  self.contents[begin:end]
                begin=r.offsets[-1][1]
            elif (r.index > 0):
                next_offset = None
                if( len(self.records) > (r.index +1)):
                    for ro in self.records[r.index+1:]:
                        if( r.file != file ):
                            continue
                        if len(ro.offsets):
                            next_offset=ro.offsets
                            break
                if(next_offset == None) and (begin < len(self.contents)):
                    if ( last_exists):                        
                        output +=  self.contents[begin:]
                        if self.contents[begin:].rfind('\n') == -1:
                            output += '\n'
                    begin = len(self.contents)
                elif( next_offset != None ):
                    end=next_offset[0][0]
                    if ( last_exists): 
                        output +=  self.contents[begin:end]
                    begin = end
                else:
                    if (len(output)) and (output[-1] != '\n') :
                        output+='\n'
                    begin=end
            if r.has_data():
                if (r.name != None) and (last_name != None) and (str(r.name) == last_name) and (len(r.offsets) <= 0):
                    r.name = None
                    if (r.tokens != None) and  len(r.tokens):
                        r.tokens.pop(0)
                if( r.name != None ):
                    last_name=str(r.name)                
                output += r.out(origin)
                if r.Type() == 'ORIGIN':
                    origin=r.origin
                if( begin == end):
                    if (len(output)) and (output[-1] != '\n') :
                        output += "\n"
                last_exists=True
            else:
                last_exists=False
                next_end=len(self.contents)
                for i in range(r.index+1, len(self.records)):
                    if (self.records[i].file != file):
                        continue
                    if len(self.records[i].offsets):
                        begin = self.records[i].offsets[-1][-1]
                        for j in range(i, len(self.records)):
                            if len(self.records[j].offsets):
                                next_end = self.records[j].offsets[0][0]
                                end = next_end
                begin = end
                if( len(self.contents) > begin ):
                    while ( begin <= end ) and (self.contents[begin] != '\n'):
                        begin-=1
        if self.records[last_index].has_data() and len(self.records[last_index].offsets):
            output +=self.contents[self.records[last_index].offsets[-1][1]:]
        elif(output[-1] != '\n'):
            output += '\n'
        if (file != None):
            if (len(output)==0):
                output=self.files[file]
            self.contents=saved_contents
        return output

    def save(self):
        d=datetime.datetime.now()
        ts=str(d.year)+'-'+str(d.month)+'-'+str(d.day)+'_'+str(d.hour)+'.'+str(d.minute)+'.'+str(d.second)
        self.errors=[]
        debug( 'SAVE ZONE:',self.zone_origin,self.cfgType,self.filename)
        if(self.filename == None) or (self.cfgType == 'slave'):
            return
        if hasattr(self,'SOA') and hasattr(self.SOA,'SERIAL'):
            self.SOA.SERIAL += 1
        try:
            backedUp = False
            if os.access(self.filename, os.F_OK) == 1:
                os.rename(self.filename, self.filename+'.'+ts)
                backedUp = True
            fd=os.open(self.filename, os.O_WRONLY | os.O_CREAT, 0644)            
            os.write(fd, self.out())
            os.fsync(fd)
            os.close(fd)
            commands.getstatusoutput( "if [ -x /sbin/restorecon ]; then /sbin/restorecon "+ self.filename + "; fi" )
            for f in self.files.keys():
                os.rename(f, f+'.'+ts)
                fd=os.open(f, os.O_WRONLY | os.O_CREAT, 0644)                
                os.write(fd,self.out(f))
                os.fsync(fd)
                os.close(fd)
                commands.getstatusoutput( "if [ -x /sbin/restorecon ]; then /sbin/restorecon "+ f + "; fi;" )
            s=commands.getstatusoutput('/usr/sbin/named-checkzone ' + self.origin + ' ' + self.filename)
            if( s[0]!= 0) and len(s[1]):
                self.errors.append(_('Save of Zone File %s failed:\n%s.')%(self.filename,s[1]))
                if  backedUp :
                    os.rename(self.filename, self.filename+'.REJECT'+'.'+ts)
                    os.rename(self.filename+'.'+ts, self.filename)
                    for f in self.files.keys():
                        os.rename(f, f+'.REJECT'+'.'+ts)
                        os.rename(f+'.'+ts, f)
        except OSError, e:
            self.errors.append(_('Save of Zone File %s failed:\n%s.')%(self.filename,str(e[0])+':'+e[1]))
            return
        self.additions = 0

    def preview(self):
        yield ( self.filename, self.out() )
        for f in self.files.keys():
            yield ( f, self.out(f) )
    
    def __init__(self, filename, origin, nmdc, nmdcKey, clas='IN', cfgType='master', continuation=None):
        self.filename=filename
        self.nmdc=nmdc
        self.nmdcKey=nmdcKey
        self.records=[]
        self.loaded = False
        self.contents=''
        if clas == None:
            self.clas = 'IN'     # named syntax: class is 'IN' if unspecified
        else:
            self.clas = clas
        self.clas=clas
        self.cfgType=cfgType
        if(origin[-1] != '.'):
            origin += '.'
        self.origin=origin       # current origin within zone
        self.zone_origin=origin  # initial origin of whole zone
        if ( continuation != None ):
            self.continuation=continuation
        else:
            self.continuation=False
        if(origin[-len(DNS.IPV4.REVERSE_ZONE):].upper()==DNS.IPV4.REVERSE_ZONE):
            self.type=DNS.IPV4.REVERSE
            self.origin_name = DNS.IPV4.ReverseAddress(origin,self)
        elif(origin[-len(DNS.IPV6.REVERSE_ARPA_ZONE):].upper()==DNS.IPV6.REVERSE_ARPA_ZONE):
            self.type=DNS.IPV6.REVERSE_ARPA
            self.origin_name = DNS.IPV6.ReverseARPAddress(origin,self)
        elif(origin[-len(DNS.IPV6.REVERSE_INT_ZONE):].upper()==DNS.IPV6.REVERSE_INT_ZONE):
            self.type=DNS.IPV6.REVERSE_INT
            self.origin_name = DNS.IPV6.ReverseINTAddress(origin,self)
        elif(origin[-len(DNS.NSAP.REVERSE_ZONE):].upper()==DNS.NSAP.REVERSE_ZONE):
            self.type=DNS.NSAP.REVERSE
            self.origin_name = DNS.NSAP.ReverseAddress(origin,self)
        else:
            self.type=DNS.FORWARD
            self.origin_name = DNS.Name(origin,self)
#        print 'TYPE: ',self.type
#        print 'ZONE  ORIGIN: ',self.zone_origin, 'TYPE:', self.type
        if( self.filename != None):
            self.load(continuation==None)
        self.loaded = True
        
    def dump(z):
        for r in z.records:
            debug( r.index, r.__class__)
            debug( r.Name(),':' )
            if not r.has_data():
                debug('DELETED!')
            if not r.is_directive():
                for name, val in r.value_labels():
                    debug( name, '=', val)
            else:
                debug( '$'+r.Type(),'=',r.value())
            debug( '-')
    
#             for k in r.keys():
##                     if (k == 'name') and (r[k] == None) :
##                          n=r.Name()
##                          print ' ',k,'=',type_or_instance(r[k]), r[k], '(', n ,')'
##                          print 'label: ', n.label()
##                          print 'relative label: ', n.relative_label()
##                      elif(k == 'clas') and (r[k] == None) :
##                          print ' ',k,'=',type_or_instance(r[k]), r[k], '(', r.Class() ,')'
##                      elif(k == 'ttl') and (r[k] == None) :
##                          print ' ',k,'=',type_or_instance(r[k]), r[k], '(', r.TTL() ,')'
##                      else:
##                          print ' ',k,'=',type_or_instance(r[k]), r[k]
    def origin_label(z):
        if hasattr(z, 'SOA'):
            return z.SOA.Name().label()
        else:
            if z.type == DNS.FORWARD:
                return z.zone_origin
            elif z.type == DNS.IPV4.REVERSE:
                n = DNS.IPV4.ReverseAddress(z.zone_origin,z)
                return n.label()
            elif (z.type == DNS.IPV6.REVERSE_ARPA) or (z.type == DNS.IPV6.REVERSE_INT):
                n = DNS.IPV6.ReverseARPAddress(z.zone_origin,z)
                return n.label()
            elif z.type == DNS.NSAP.REVERSE:
                n = DNS.NSAP.ReverseAddress(z.zone_origin,z)
                return n.label()
            else:
                return z.zone_origin
# END class Zone

def type_or_instance(o):
    if type(o) == types.InstanceType:
        return o.__class__
    return str(type(o))
        
class NullNmdc:
    options={ 'directory': '/var/named' }
    def directory(s):
        return '/var/named'

if __name__ == "__main__":    
    if (len(sys.argv)>2) and len(sys.argv[1]) and len(sys.argv[2]) :            
        z=Zone(sys.argv[1], sys.argv[2], NullNmdc(), ['zone',sys.argv[2]])
#        z.records[-1].tokens=[]
#        z.records[-2].tokens=[]
#        z.records[0].tokens=[]
#        debug( 'NEW RR jvdlab')
#        name=DNS.Name('jvdlab',z)
#        debug( 'NEW RR:',str(name),name.value())
#        rr=RR.AAAA(z,name,None,None,None,None)
#        debug( 'RR NAME:',rr.name)
#        rr.ADDRESS=DNS.IPV6.AAAA('2001:1:2:3:4:5:6:7')
#        rr.update()
#        z.dump()
#        z.records[11].tokens=[]
#        r=RR.CNAME(z, DNS.Name('test8',z), z.clas, z.ttl, None, None)
#        r.CNAME=DNS.Name('jvdlab.jvd.boston.redhat.com.',z)
#        r.update()
#        z.records[19].tokens=[]
#        r=RR.A(z, DNS.Name('test16',z),z.clas,z.ttl,None,None)
#        r.ADDRESS=DNS.IPV4.Address('64.32.16.17')
#        r.update()
#        z.records[13].tokens=[]
#        z.records[14].tokens=[]
#        z.records[16].tokens=[]
#        r=RR.A(z, DNS.Name('test13.fuckup.net.',z),z.clas,z.ttl,None,None)
#        r.ADDRESS=DNS.IPV4.Address('64.32.16.18')
#        r.update()
#        r=RR.CNAME(z, DNS.Name('test12',z), z.clas, z.ttl, None, None)
#        r.CNAME=DNS.Name('jvdlab.jvd.boston.redhat.com.',z)
#        r.update()
#        z.dump()        
        r=RR.A(z, DNS.Name('test17',z),z.clas,z.ttl,None,None)
        r.ADDRESS=DNS.IPV4.Address('64.32.16.17')
        r.update()
        os.write(1,z.out())
#        key = RR.DNSKEY( z, DNS.Name(z.zone_origin, z), 'IN', None,
#                      ['jvd.boston.redhat.com.', 'IN', 'DNSKEY',
#                       '257', '3', '5', 'AQO6dUxrHll9HDTxzE8uOXzlPIfbAvpJOjH3A59HhI7LdmkFblE7lpwH',
#                       '6g6xOfBN16A7','tDKJ+hIlWRixUfv8Pul/VGuf4GaPZkrtQLrIwE6LnPGC',
#                       'KlMbXUgqkWTFAYn7ScY92mz69nqZfKQrZYyulQAppygHTyY5QSggm/Bk', '4SU8sQ=='
#                      ],[], 3 )
#        key.update()
#        rr=RR.A(z, DNS.Name('jvdlab',z), z.clas, z.ttl, None, None)
#        rr.ADDRESS = DNS.IPV4.Address('172.16.80.221')
#        rr.update()
#        rr=RR.A(z, DNS.Name('jvdias',z), z.clas, z.ttl, None, None)
#        rr.ADDRESS = DNS.IPV4.Address('172.16.80.119')
#        rr.update()
#        for r in z.records:
#            print '-'
#            print r.index
#            print 'TOKENS:', r.tokens
#            print 'OFFSETS:', r.offsets
#        r=RR.PTR(z,DNS.IPV4.ReverseAddress('172.16.80.221',z),z.clas,z.ttl,None,None)
#        r.PTRDNAME = DNS.Name('jvdlab.jvd.boston.redhat.com.',z)
#        r.update()
#        r=RR.PTR(z,DNS.IPV4.ReverseAddress('172.16.80.119',z),z.clas,z.ttl,None,None)
#        r.PTRDNAME = DNS.Name('jvdias.jvd.boston.redhat.com.',z)        
#        r.update()
#        z.records[30].tokens=[]
#        os.write(1,z.out())
#        print z.records
#        del z.records[8].tokens
#        newRR=RR.A(z,DNS.Name("jvdsppc",z),"IN",None,["A", "172.16.80.39"],[])        
## #        z.records[1].ttl='1h'
## #        z.records[3].clas='IN'
## #zone,name,clas,ttl,tokens,offsets):
##         os.write(1,z.out())
##         for f in z.files.keys():
##             print '---',f,': ---'
##             os.write(1, z.out(f))
#        z.save()

