import sys
import os
import time
import signal
import gtk
import gtk.glade
import gnome.ui
import gnome
import gobject
import pango

import DNS
import RRGUI
from Zone   import *
from View   import *
from RREdit import *
from EditDialog import *
from NewZone import *
from NewView import *
from Import  import *
from RR     import *
from Server import *
import DNSsec

PIX_COLUMN=0
TEXT_COLUMN=1
INFO_COLUMN=2
RECORD_COLUMN=3

RDNC_KEY_NAME=_('Remote Nameserver Control Security Key')

class ZoneNode:
    def __init__(s, z):
        s.z=z
    
    def __hasattr__(s, attr):
        return s.z.has_key(attr)
    
    def __getattr__(s, attr):
        if attr == 'z':
            return s.z
        if s.z.has_key(attr):
            return s.z[attr]
        return None    
        
    def __delattr__(s, attr):
        if s.z.has_key(attr):
            del s.z[attr]
        
    def __getitem__(s, key):
        return s.__getattr__(key)

    def __setitem__(s, key, val):
        s.z[key]=val

    def __delitem__(s, key):
        s.__delattr__(key)
    
    def __contains__(s,key):
        return s.__hasattr__(key)
    
    def keys(s):
        for k in s.z.keys():
            yield k
    
    def __str__(s):
        return s.z['zone'].zone_origin

 
class NewACL:
    def response(s,d,r):
        if r == gtk.RESPONSE_OK :
            if not(hasattr(s,'name')):                
                s.name_check(s,s.entry)
                if not(hasattr(s,'name')):                
                    return False
            s.acl.key = s.nmdcKey + [s.name]
            s.acl.name = s.name
            s.aclWidget.save()
            s.zoneTree.load()
        s.dialog.destroy()
        return False
    
    def name_check(s,w,e=None):
        name=s.entry.get_text()
        if (len(name)==0):
            s.bind.gui.modalError(s.dialog,_("ACL Name Error"),
                                  _('You must enter a valid ACL name.')
                                 )
            return False
        if (DNS.ACL.namedACLs.has_key(name)):
            s.bind.gui.modalError(s.dialog,_("ACL Name Error"),
                                  _('An ACL with this name already exists.')
                                 )
            return False
        if s.bind.keys.has_key(name):
                s.bind.gui.modalError(s.dialog,_("ACL Name Error"),
                                             _('A security key with this name already exists.')
                                            )
                return False
        if name in [ 'none', 'any', 'localhost', 'localnets' ]:
                s.bind.gui.modalError(s.dialog,_("ACL Name Error"),
                                             _('The ACL name cannot be an ACL name keyword.')
                                            )
                return False
        s.name = name
        return False
        
    def __init__(s,selection,zoneTree):
        s.nmdcKey=['acl']
        if selection != None:
            s.nmdcKey = selection.nmdcKey + ['acl']
        debug( 'NEW ACL:',s.nmdcKey)
        s.bind = zoneTree.bind
        s.gui=s.bind.gui
        s.zoneTree = zoneTree
        s.nmdc = s.bind.nmdc
        s.acl=DNS.ACL(None,s.nmdc)
        s.dialog = gtk.Dialog(title=_("New Access Control List"), parent=s.bind.gui.gui,
                              buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                      )
                             )
        s.dialog.connect('response',s.response)
        s.entry = gtk.Entry()
        s.entry.connect('focus_out_event',s.name_check)
        s.nameFrame = gtk.Frame(label=_('Access Control List Name'))
        s.nameFrame.add(s.entry)
        s.dialog.vbox.pack_start(s.nameFrame,False)
        s.acl.name=''
        s.aclWidget = RRGUI.DNS_ACL(s,s,s.acl,'acl')
        s.dialog.vbox.pack_start(s.aclWidget.widget())
        s.dialog.show_all()

class NewServer:
    def response(s,d,r):
        if r == gtk.RESPONSE_OK:
            s.addressWidget.save()
            addr = str(s.address)
            if len(addr) == 0:
                s.bind.gui.modalError(s.dialog,_('Server Address Error'), _('You must enter a valid server address.'))
                return False
            if s.bind.nmdc.__contains__(s.nmdcKey + [addr]):
                 s.bind.gui.modalError(s.dialog,_('Server Address Error'),
                                       _('Server %s already exists. Specify another address.') % addr
                                      )
                 return False
            name = s.nmdcKey
            if s.bind.nmdc.__contains__(s.nmdcKey):
                name = s.nmdcKey + [ addr ]
            s.bind.nmdc[s.nmdcKey] = 'server '+addr+' '+'{\n};\n'
            s.bind.nmdc.update()
            if s.nmdcKey[0] == 'view':
                s.bind.views[s.nmdcKey[0].strip('"')]['servers'][addr]=Server( s.bind.nmdc, s.nmdcKey+[addr])
            else:
                s.bind.servers[addr]=Server( s.bind.nmdc,s.nmdcKey+[addr])
            s.zoneTree.load()
        d.destroy()
        
    def __init__(s,zoneTree,parent):
        s.zoneTree = zoneTree
        if (type(parent) == type(None)) or (parent.__class__ != View):
            s.nmdcKey = ['server']
        else:
            s.nmdcKey = parent.nmdcKey[:] + [ 'server' ]
        debug( 'NEW SERVER:',s.nmdcKey)
        s.dialog = gtk.Dialog(title=_("New Server"), parent=s.zoneTree.bind.gui.gui,
                              buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                      )
                             )
        s.gui = s.zoneTree.gui
        s.bind= s.zoneTree.bind
        s.dialog.connect('response',s.response)
        s.address=DNS.IPV4.Address('0.0.0.0')
        s.addressWidget = RRGUI.DNS_Address(s,s,s,'address')
        s.aFrame=gtk.Frame(_('Server Address'))
        s.aFrame.add(s.addressWidget.widget())
        s.dialog.vbox.pack_start(s.aFrame)
        s.dialog.show_all()

class AddMenu:
    def selected(s, i, r):
        debug( 'NEW',s.selection, r)
        if issubclass(r,RR):
            rr = r(s.zone,'',s.zone.clas,s.zone.ttl,None,None)
            RREdit(s.zoneTree, rr)            
        elif r == Zone:
            if ( s.selection.__class__ == View ):
                NewZone(s.zoneTree,s.selection.nmdcKey[:])
            else:
                NewZone(s.zoneTree)
        elif r == View:
            NewView(s.zoneTree.bind)
        elif r == DNS.ACL:
            NewACL(None, s.zoneTree)
        elif r == DNS.Key:
            key = [ 'key' ]
            if s.selection != None:
                key = s.selection.nmdcKey + key
            k = DNS.Key(None,s.zoneTree.bind.nmdc, key)
            RRGUI.DNS_Key(s.zoneTree,s,k,key).keyDialog()
        elif r == DNS.Controls:
            RRGUI.DNS_Controls(s,s,DNS.Controls(s.bind.nmdc),'controls').ctlDialog()
        elif r == DNS.Logging:
            RRGUI.DNS_Logging(s,s,DNS.Logging(s.bind.nmdc),'logging').loggingDialog()
        elif r == DNSsec.TrustedKeys:
            r(s.zoneTree.bind).EditDialog()
        elif r == Server:
            NewServer(s.zoneTree,s.selection)
            
    def detach(s):
        debug( 'DETACH')
        return False
    
    def __init__(s, z, selection, widget=None):
        s.selection = selection
        s.zoneTree = z
        s.bind=z.bind
        s.gui=s.bind.gui
        s.menu = gtk.Menu()
        choices=[]
        show_names=False
        max_name_len=0
        if( type(selection) == type(None) ):
            if s.zoneTree.bind.isViewed:
                choices=[]
            else:
                choices=[['zone',_('Zone'),Zone]]
            choices+=[['view',_('View'),View],
                      ['acl', _('Access Control List'), DNS.ACL],
                      ['key',_('Security Key'), DNS.Key],
                      ['server',_('Server'),Server]
                     ]
            if not s.bind.nmdc.__contains__('controls'):
                choices.append(['controls', _('Controls'), DNS.Controls])
            if not s.bind.nmdc.__contains__('logging'):
                choices.append( ['logging', _('Logging'), DNS.Logging] )
            if not s.bind.nmdc.__contains__('trusted-keys'):
                choices.append( ['trusted-keys', _('DNSSEC Trusted Keys'), DNSsec.TrustedKeys ] )
        elif(selection.__class__ == View):
            choices=[['zone',_('Zone'),Zone],
                     ['key',_('Security Key'), DNS.Key]
                    ]
        elif(selection.__class__ == ZoneNode ) or  (isinstance(selection,RR)):
            rrs=[]
            s.zone = selection.zone
            for rn in RDATA.keys():
                rr = RDATA[rn]
                if hasattr(rr,'directive') or hasattr(rr,'deprecated') or hasattr(rr,'todo'):
                    continue
                if hasattr(rr,'unique') and hasattr(s.zone,rn):
                    continue # SOA
                if not s.zone.type in rr.types:
                    continue
                if hasattr(rr,'dnssec'):
                    continue
                rrs.append(rn)
            rrs.sort()
            for rn in rrs:
                rr = RDATA[rn]
                choices.append([rn, rr.desc[''], rr])
                if len(rn) > max_name_len:
                    max_name_len = len(rn)                    
            show_names = True
        elif((selection.__class__ == DNS.ACL) or (selection.__class__==DNS.Key)):
            choices=[['acl',_('Access Control List Filter'),DNS.ACL],
                     ['key', RDNC_KEY_NAME, DNS.Key]
                    ]
        for choice in choices:
            if show_names :
                name =( ('%-' + str(max_name_len+1) + 's') % choice[0] ) + choice[1]
            else:
                name = choice[1]
            item = gtk.MenuItem(label=name)
            item.connect('activate',s.selected, choice[2])
            s.menu.append(item)
        if( widget != None ):
            s.menu.attach_to_widget(widget,s.detach)
            s.menu.popup(None,None,None,0,0)
            s.menu.show_all()

class RootCacheDownload:
    def root_cache_download(s):
        gobject.source_remove(s.rcdSID)
        r = s.zoneTree.bind.get_root_hints(filename=s.selection['config']['file'].strip('"'),
                                           dir=s.zoneTree.bind.nmdc.directory()
                                          )
        if r != None:
            s.rcd.dialog.destroy()
            s.zoneTree.bind.gui.modalError(s.zoneTree.bind.gui.gui,_('Unable to download the root hints cache.'),r)
        else:
            s.rcd.errorText.set_text('The root hints cache was downloaded successfully.')
    
    def get_root_cache(s):
        s.rcd = s.zoneTree.bind.gui.info(s.zoneTree.bind.gui.gui,'','')
        s.rcd.errorTitle.set_markup(_('Root Cache Update'))
        s.rcd.errorText.set_text(_('Download in Progress'))
        s.rcd.dialog.resize_children()
        s.rcd.dialog.show_all()
        s.rcdSID=gobject.timeout_add(190,s.root_cache_download)
    
    def __init__(s,zoneTree, selection):
        s.zoneTree=zoneTree
        s.selection=selection
        s.get_root_cache()

class ServerControl:    
    def doIt(s,action):
        gobject.source_remove(s.sid)
        action(1)
        
    def start(s,cb=None):        
        if cb==None:
            s.show_dialog()
            s.dialog.errorText.set_text(_('Starting the DNS Server...'))
            s.sid=gobject.timeout_add(200,s.doIt,s.start)
        else:
            r = s.bind.start_named()
            if r[0] != 0:
                s.dialog.errorText.set_text(_('Failure:') + '\n' + r[1])
            else:
                s.dialog.errorText.set_text(_('Server start succeeded.'))
        s.dialog.dialog.resize_children()
        s.dialog.dialog.show_all()
    
    def stop(s,cb=None):
        if cb==None:
            s.show_dialog()
            s.dialog.errorText.set_text(_('Stopping the DNS Server...'))
            s.sid=gobject.timeout_add(200,s.doIt,s.stop)
        else:
            r=s.bind.stop_named()
            if r[0] != 0:
                s.dialog.errorText.set_text(_('Failure:\n %s') % r[1])
            else:
                s.dialog.errorText.set_text(_('Server shutdown succeeded.'))                
        s.dialog.dialog.resize_children()
        s.dialog.dialog.show_all()
       
    def restart(s,cb=None):
        if cb==None:
            s.show_dialog()
            s.dialog.errorText.set_text(_('Restarting the DNS Server...'))
            s.sid=gobject.timeout_add(200,s.doIt,s.restart)
        else:
            r=s.bind.restart_named()
            if r[0] != 0:
                s.dialog.errorText.set_text(_('Failure:\n %s') % r[1])
            else:
                s.dialog.errorText.set_text(_('Server restart succeeded.'))                
        s.dialog.dialog.resize_children()
        s.dialog.dialog.show_all()                
    
    def reload(s,cb=None):
        if cb==None:
            s.show_dialog()
            s.dialog.errorText.set_text(_('Restarting the DNS Server...'))
            s.sid=gobject.timeout_add(200,s.doIt,s.reload)
        else:
            r=s.bind.reload_named()
            if r[0] != 0:
                s.dialog.errorText.set_text(_('Failure:\n %s') % r[1])
            else:
                s.dialog.errorText.set_text(_('Server reload succeeded.'))
        s.dialog.dialog.resize_children()
        s.dialog.dialog.show_all()
    
    def show_dialog(s):
        s.dialog = bind.gui.info(bind.gui.gui,'','')
        s.dialog.errorTitle.set_markup(_('DNS Server Control'))
        s.dialog.dialog.resize_children()
        
    def options(s):
        if s.bind.named_is_running():
            yield (_('Restart Server'), s.restart)
            yield (_('Reload Server'), s.reload)
            yield (_('Stop Server'), s.stop)
        else:
            yield ( _('Start Server'),s.start )
    
    def __init__(s,bind):
        s.bind = bind

class PopupMenu:
    def add(s, arg):
        debug( 'ADD', arg)
    
    def detach(s):
        debug( 'DETACH')
        return False

    def edit(s, i):
        debug( 'EDIT', s.selection)
        s.zoneTree.Edit(s.selection)
        return False

    def delete(s, i):
        debug( 'DELETE', s.selection)
        if type(s.selection) != type(None):
            Delete(s.zoneTree, s.selection)
        return False

    def get_root_cache(s,i):
        RootCacheDownload(s.zoneTree,s.selection)

    def controlServer(s,i,a):
        a()

    def signZoneResponse(s,answer,noparms=None):
        if (answer==gtk.RESPONSE_OK):
            bind.save()    
            DNSsec.SignZone(s.zoneTree.bind, s.selection.zone)
            return True 

    def signZone(s,i):
        if (bind.dirty()):
            ask = bind.gui.question(None, _('DNS Configuration Modified'), _('The DNS Configuration has to be saved before signing zone! Do you want to proceed?'))
            ask.question(s.signZoneResponse)
        else:    
            DNSsec.SignZone(s.zoneTree.bind, s.selection.zone)
    
    def __init__(s, z, w, event, selection):
        s.selection = selection
        s.zoneTree = z
        s.menu = gtk.Menu()        
        s.menu.attach_to_widget(w, s.detach)
        masterZone=False
        if not(   ( (s.selection.__class__==ZoneNode) and (s.selection['zone'].cfgType in ['forward','slave','stub','hint']))
               or ( isinstance(s.selection,RR) and (s.selection.zone.cfgType in ['slave','stub','forward','hint']) )
              ):
            s.addItem = gtk.MenuItem(label=_('Add...'))
#        s.addItem.connect('activate',s.add)        
            s.addMenu = AddMenu(z, selection )
            s.addItem.set_submenu(s.addMenu.menu)
            s.menu.append(s.addItem)
            s.delItem = gtk.MenuItem(label=_('Delete...'))
            s.delItem.connect('activate',s.delete)
            s.menu.append(s.delItem)
            masterZone = (s.selection.__class__==ZoneNode)
        elif((s.selection.__class__==ZoneNode) and (s.selection['zone'].cfgType == 'hint')):
            s.dnldItem = gtk.MenuItem(label=_('Update Root Cache'))
            s.dnldItem.connect('activate',s.get_root_cache)
            s.menu.append(s.dnldItem)
        s.editItem = gtk.MenuItem(label=_('Edit...'))
        s.editItem.connect('activate',s.edit)
        s.menu.append(s.editItem)
        if type(s.selection) == type(None):
            s.menu.append(gtk.SeparatorMenuItem())
            s.serverCtl = ServerControl(s.zoneTree.bind)
            for desc, act in s.serverCtl.options():
                i = gtk.MenuItem(label=desc)
                i.connect('activate',s.controlServer,act)                    
                s.menu.append(i)
        if masterZone and ((s.selection.dnssecSecure == None) or (s.selection.dnssecSecure==False)):
            s.menu.append(gtk.SeparatorMenuItem())
            s.signZoneItem = gtk.MenuItem(label=_('Sign...'))
            s.signZoneItem.connect('activate',s.signZone)
            s.menu.append(s.signZoneItem)
        s.menu.popup(None, None, None, 0, 0)    
        s.menu.show_all()

class Save:
    def __init__(s,bind,onExit=False):
        s.bind = bind
        s.bind.onExit = onExit
        if not bind.dirty():
            s.bind.gui.warning(s.bind.gui.gui, _('DNS Configuration Warning'), _('There are no changes to save.'))
        else:            
            s.dialog = s.bind.gui.question(s.bind.gui.gui, _('DNS Configuration Save'), _('Do you really want to backup and replace the existing DNS configuration files?'))
            s.dialog.question(s.answer)        
            
    def answer(s, response, noparms=None):
        if( response == gtk.RESPONSE_OK ):
            if s.bind.save() and s.bind.named_is_running():                
                r = s.bind.restart_named()
                if r[0] != 0:
                    s.bind.gui.modalError(s.bind.gui.gui,_('named restart failed:'),_(r[1]))
                    return
                else:
                    s.bind.gui.warning(s.bind.gui.gui,_('Information'),_('DNS configuration was saved and named was restarted successfully.'))
        if s.bind.onExit:
            s.bind.gui.exit(answer=gtk.RESPONSE_OK)            
            
class Delete:
    def response(s, r, p=None):
        if r == gtk.RESPONSE_OK:
            if issubclass(s.type,RR):
                debug( 'DELETE', s.obj.zone.zone_origin, s.obj.index)
                s.obj.tokens=[]
                if hasattr(s.obj,'new_record'):
                    del s.obj.new_record
                if s.obj.name != None:
                    if( (s.obj.index < (len(s.obj.zone.records)-1)) and
                        (s.obj.zone.records[ s.obj.index ].name != None) and
                        (s.obj.zone.records[ s.obj.index + 1 ].name == None)
                      ):
                        s.obj.zone.records[ s.obj.index + 1 ].name = s.obj.name
                zn = s.zoneTree.bind.get_zone_node(s.obj.zone)
                if zn != None:
                    zn['dirty'] = True
                    debug( 'SET DIRTY OK')
                s.zoneTree.load()
            elif hasattr(s.obj,'nmdcKey'):                
                debug( 'DELETE', s.obj.nmdcKey)
                del s.zoneTree.bind.nmdc[s.obj.nmdcKey]
                s.zoneTree.bind.nmdc.update()
                s.zoneTree.bind.load()
            elif hasattr(s.obj,'key'):
                debug( 'DELETE', s.obj.key)
                del s.zoneTree.bind.nmdc[s.obj.key]
                s.zoneTree.bind.nmdc.update()
                s.zoneTree.bind.load()
            elif s.obj.__class__ == DNS.Controls:
                debug( 'DELETE Controls')
                del s.zoneTree.bind.nmdc['controls']
                s.zoneTree.bind.nmdc.update()
                s.zoneTree.bind.load()
            elif s.obj.__class__ == DNS.Logging:
                debug( 'DELETE Logging')
                del s.zoneTree.bind.nmdc['logging']
                s.zoneTree.bind.nmdc.update()
                s.zoneTree.bind.load()
            elif s.obj.__class__ == Server:
                debug( 'DELETE Server ', str(s.obj.address) )
                del s.zoneTree.bind.nmdc[['server',str(s.obj.address)]]
                s.zoneTree.bind.nmdc.update()
                s.zoneTree.bind.load()
        return False
    
    def __init__(s, zoneTree, obj):
        s.zoneTree = zoneTree
        s.obj = obj
        if s.obj.__class__==ZoneNode:
            s.type=Zone
            s.obj = s.obj['zone']
        else:
            s.type=s.obj.__class__
        s.container=''
        debug( 'DELETE: ',s.type)
        if s.type == Zone:
            s.name = _('Zone')+' '+s.obj.zone_origin
            s.container=_('and all its contents')
        elif s.type == View:
            s.name = _('View')+' '+s.obj.v['view']
            s.container=_('and all its contents')
        elif s.type == DNS.ACL:
            debug( 'delete ACL:',s.obj.name, s.obj.key)
            if s.obj.name != None:
                acl_users = s.zoneTree.bind.acl_users(s.obj.name)
                if len(acl_users) != 0:
                    s.zoneTree.bind.gui.modalError(s.zoneTree.bind.gui.gui,_("ACL Delete Error"),
                                                   _("This Access Control List cannot be deleted because\nit is used by other elements:\n%s.") % string.join(acl_users,"\n"))
                    return
                s.name = 'ACL ' + s.obj.name                
            else:
                s.name = 'ACL ' + string.join( s.obj.key,' ')
            debug( 'ACL NAME:',s.name)
        elif s.type == DNS.Key:
            s.name = s.obj.name
            if (s.name == None) or (len(s.name) <= 0):
                if len(s.obj.key) > 0:
                    s.name = s.obj.key[1]
                else:
                    s.name = _('this key')
        elif s.type == DNS.Controls:
            s.name = _('the DNS Server Controls')
        elif s.type == DNS.Logging:
            s.name = _('the DNS Server Logging Configuration')
        elif s.type == Server:
            s.name = str(s.obj.address)
        elif issubclass(s.type,RR):
            s.name = obj.Type() + ' ' + str(obj.Name())
        else:
            s.name = _('this')
        if ( s.name == None ) or (len(s.name)==0):
            s.name = _('this')
        ask=s.zoneTree.bind.gui.question(s.zoneTree.bind.gui.gui,_('Confirm Delete'),_('Do you really want to delete %s\n%s?\nWARNING: this action cannot be undone.') % (s.name, s.container)
                                        )
        ask.question(s.response)
    
class Preview:
    def PreviewWindowOK(s, w, r):
        w.destroy()

    def PreviewWindow(s, i, pv):
        name,contents=pv
        td=gtk.Dialog(title=name,parent=s.bind.gui.gui,buttons=(gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        sw=gtk.ScrolledWindow()
        cn=gtk.Label(contents)
        af=pango.AttrFamily('Monospace',0,len(contents))
        al=pango.AttrList()
        al.insert(af)
        cn.set_attributes(al)
        sw.add_with_viewport(cn)
        td.vbox.pack_start(sw)
        td.connect('response', s.PreviewWindowOK)
        td.set_size_request(400,400)
        td.show_all()
        
    def __init__(s, bind):
        s.bind = bind
        s.previews=[]
        s.menu = gtk.Menu()
        dir = s.bind.nmdc.chroot        
        for file, contents in s.bind.nmdc.preview():
            if (len(file) > len(dir)) and (file[:len(dir)] == dir):
                file = file[len(dir):]
            name = _('Named Configuration %s') % file
            item = gtk.MenuItem(label=name)
            item.connect('activate',s.PreviewWindow,(name, contents))
            s.menu.append(item)
        dir=s.bind.nmdc.directory()
        for zk in s.bind.zones.keys():
            zs=[]
            if type( s.bind.zones[zk] ) == list:
                zs =  s.bind.zones[zk]
            else:
                zs = [ s.bind.zones[zk] ]
            for z in zs:
                for file, contents in z['zone'].preview():
                    if file == None:
                        continue
                    if (len(file)>len(dir)) and (file[:len(dir)] == dir):
                        file = file[len(dir)+1:]
                    name = _('Zone')+' '+z['zone'].zone_origin + ' ' + file
                    item = gtk.MenuItem(label=name)
                    item.connect('activate',s.PreviewWindow,(name, contents))                        
                    s.menu.append(item)
        s.menu.popup(None,None,None,0,0)
        s.menu.show_all()

class DragZone:
    
    def copy(s,i):
        zone = s.zoneTree.store.get_value(s.zoneIter, RECORD_COLUMN)
        view = s.zoneTree.store.get_value(s.viewIter, RECORD_COLUMN)
        oldKey = zone['parent']+ [zone['key']]
        if s.nmdc[['view',view.v['key']]].has_key('zone') and s.nmdc[['view',view.v['key'],'zone']].has_key(zone['key']):
            s.zoneTree.bind.gui.modalError(s.zoneTree.bind.gui.gui,_('Zone Drag Error'),_('Zone %s already exists in view %s.') % ( zone['zone'].zone_origin, view.key.strip('"')) )
            return False
        if not s.nmdc[['view',view.v['key']]].has_key('zone'):
            newKey = [ 'view', view.v['key'], 'zone' ]
        else:
            newKey = [ 'view', view.v['key'], 'zone', zone['key'] ]
        zoneText = s.nmdc.out(oldKey)
        s.nmdc[ newKey ] = zoneText
        if( i != s.moveItem ):
            s.nmdc.update()
            f = s.nmdc[ ['view', view.v['key'], 'zone', zone['key'], 'file'] ]
            f = f.strip('"')
            old_file=s.nmdc.directory() + '/' + f
            f =  view.v['key'].strip('"')  + '_' + f
            if os.access( s.nmdc.directory() + '/' + f, os.F_OK )==1:
                try:
                    os.rename(s.nmdc.directory() + '/' + f, s.nmdc.directory() + '/' + f + '.' + str(time.time()))
                except OSError,e:
                    debug( 'ERROR on rename', s.nmdc.directory() + '/' + f,' to ',s.nmdc.directory() + '/' + f + '.' + str(time.time()),':', e)
            try:
                out_fd=os.open(s.nmdc.directory() + '/' + f,os.O_WRONLY | os.O_CREAT)
                in_fd=os.open(old_file,os.O_RDONLY)
                r=os.read(in_fd,512)
                while len(r):
                    os.write(out_fd,r)
                    r=os.read(in_fd,512)
                os.fsync(out_fd)
                os.close(out_fd)
                os.close(in_fd)
            except OSError, e:
                debug( 'ERROR copy ', old_file, ' to ',s.nmdc.directory() + '/' + f)
            s.nmdc[ [ 'view', view.v['key'], 'zone', zone['key'], 'file'] ]= '"' + f + '"'            
            s.bind.load()
        else:
            return oldKey
        return False
    
    def move(s,i):
        oldKey = s.copy(i)
        if oldKey == False:
            return oldKey
        del s.nmdc[ oldKey ]
        s.nmdc.update()
        s.bind.load()
        return False
        
    def detach(w):
        pass
    
    def __init__(s, zoneTree, zoneIter, viewIter):
        s.zoneTree = zoneTree
        s.zoneIter = zoneIter
        s.viewIter = viewIter
        s.nmdc = zoneTree.bind.nmdc
        s.bind = zoneTree.bind
        s.menu = gtk.Menu()
        s.menu.attach_to_widget(zoneTree.tree, s.detach)
        s.copyItem = gtk.MenuItem(label=_('Copy'))
        s.copyItem.connect('activate',s.copy)
        s.menu.append(s.copyItem)
        s.moveItem = gtk.MenuItem(label=_('Move'))
        s.moveItem.connect('activate',s.move)
        s.menu.append(s.moveItem)
        s.menu.append(gtk.MenuItem(label=_('Cancel')))
        s.menu.popup(None,None,None,0,0)
        s.menu.show_all()
        
class ZoneTree:
    def selection(s):
        selection = s.tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                r = model.get_value(selected, RECORD_COLUMN)
                s.last_path = model.get_path(selected)
                s.last_expansion = s.tree.row_expanded(s.last_path)
                return r
        return None
         
    def selected_row(s):
        selection = s.tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                s.last_path = model.get_path(selected)
                s.last_expansion = s.tree.row_expanded(s.last_path)
                return selected
        return None

    def row_record(s,selection):
        return s.store.get_value(selection, RECORD_COLUMN)
         
    def popup_menu(s,tree,event):
        if( event.button > 1 ):
            rr = s.selection()            
            s.popupMenu = PopupMenu(s,tree,event, rr)
        return False
    
    def NewClicked(s, w):
        sel=s.selection()
        debug( 'NEW Button Clicked', sel)
        if not(   ( (sel.__class__==ZoneNode) and (sel['zone'].cfgType in ['forward','slave','stub','hint']))
               or ( isinstance(sel,RR) and (sel.zone.cfgType in ['slave','stub','forward','hint']) )
            ):        
            s.addMenu = AddMenu(s, sel, s.tree)
        elif((sel.__class__==ZoneNode) and (sel['zone'].cfgType == 'hint')):
            s.popupMenu=PopupMenu(s,s.tree,None,sel)
        else:
            s.bind.gui.modalError(s.gui.gui,_('Zone Type Error'), _('You cannot add to this zone on this server.'))

    def Edit(s, r):
        readOnly = False
        roMessage= None
        if( isinstance(r,RR) and (r.zone.cfgType in ['slave','stub','forward','hint']) ):
            readOnly = True
            debug("READ ONLY ZONE:", str(r.__class__), str(r.Name()), r.zone.cfgType)  
            roMessage = _('The contents of this zone cannot be modified on this server.')
        if( isinstance(r,RR) and hasattr(r,'dnssec') ):
            readOnly = True
            roMessage = _('You cannot save DNSSEC records - use Zone->Sign.')
        debug( 'EDIT: RO:',readOnly)
        if (type(r) == type(None)) or not(isinstance(r,RR)):
            if isinstance(r,DNS.ACL):
                RRGUI.DNS_ACL(s,None,r,r.name).aclDialog(r)
            elif isinstance(r,DNS.Key):
                RRGUI.DNS_Key(s,None,r,r.name).keyDialog()
            elif isinstance(r,DNS.Controls):
                RRGUI.DNS_Controls(s,None,r,'controls').ctlDialog()
            elif isinstance(r,DNS.Logging):
                RRGUI.DNS_Logging(s,s,r,'logging').loggingDialog()
            elif isinstance(r,DNSsec.TrustedKeys):
                r.EditDialog()
            else:
                EditDialog(s.bind,r,read_only=readOnly)
        else:
            RREdit(s,r,read_only=readOnly,read_only_message=roMessage)
    
    def EditClicked(s, w):
        s.Edit( s.selection() )        
    
    def DeleteClicked(s,w):
        r=s.selection()
        debug( 'DELETE Button Clicked', r)
        if type(r) != type(None):
            Delete(s, r)
        
    def OpenClicked(s, w):
        v=s.selection()
        debug( 'IMPORT Button Clicked', v)
        if s.bind.isViewed and ((type(v) == type(None)) or (v.__class__ != View)):
            s.bind.gui.modalError(s.bind.gui.gui, _('Import Error'), _('Please select the View that is to contain imported zones.'))

        else:
            if not s.bind.isViewed :
                v = None
            Import(s,v)
        return False
        
    def PreviewClicked(s, w):
        debug( 'PREVIEW Button Clicked')
        Preview(s.bind)
        
    def SaveClicked(s, w, onExit=False):
        Save(s.bind,onExit)
        
    def on_cursor_changed(s,tree):
        r=s.selection()
        nmdcKey = None
        if(r.__class__ == ZoneNode):
            nmdcKey = r['zone'].nmdcKey
        elif(r.__class__ == DNS.Key):
            nmdcKey == r.key
        if ( (nmdcKey != None)
          and( type(nmdcKey) == list)
          and( len(nmdcKey) >= 4)
          and( nmdcKey[0] == 'view')
          and( s.n_views > 1)
           ):
            s.statusLabel.set_markup('<small>'+_('You can drag this object to another view.')+'</small>')
        else:
            s.statusLabel.set_markup('<small>'+_('Select an object; press and release the right-hand mouse button.')+'</small>')
        return False
    
    def on_row_activated(s,zoneTree, path, column):
        node = s.store.get_iter(path)
        r = s.store.get_value(node, RECORD_COLUMN)
        debug( "TREE SELECTION: ", r.__class__)
        s.Edit(r)

    def drag_begin(s,w,dc):
        r = s.selected_row()
        s.dndSource = r
        return False
    
    def drag_end(s,w,dc):
        x,y=s.tree.get_pointer()
        try:
            p,d = s.tree.get_dest_row_at_pos(x,y)
        except TypeError:
            # not dropped on any tree item
            debug( 'Not dropped on any item' )
            return
        if( p != None):
            destIter=s.store.get_iter(p)
            dest = s.row_record(destIter)
            if ( type(dest) != type(None) ):
                if (not  hasattr(s,'dndSource') ) or (s.dndSource == None):
                    debug( 'No dndSource' )
                    return False
                source=s.row_record(s.dndSource)
                if(    (source.__class__ == ZoneNode) and (dest.__class__ == View)
                   and (dest.nmdcKey != source['zone'].nmdcKey[:-(len(dest.nmdcKey))])
                  ):
                    DragZone(s,s.dndSource,destIter)
                else:
                    debug( 'BAD SOURCE:', source.__class__, ' OR DEST:', dest.__class__ )
        return False

    def set_zone_search_expansion(s, z, sv, lsv):
        zone_path = s.store.get_path(z['node'])
        expanded = s.tree.row_expanded(zone_path)        
        match = False
        changed = False
        if( lsv > 0 ):
            origin=''
            if ( z['zone'].type == DNS.FORWARD ):                
                if hasattr(z['zone'],'SOA'):
                    origin = str(z['zone'].SOA.Name())                    
                elif (len(z['zone'].records)>0):
                    origin = str(z['zone'].records[0].Name())
                else:
                    origin = z['zone'].origin
            else:
                if hasattr(z['zone'],'SOA'):
                    origin = z['zone'].SOA.Name().label()
                elif (len(z['zone'].records)>0):
                    origin = z['zone'].records[0].Name().label()
                else:
                    origin = z['zone'].origin
            if ( sv == origin ):
                if z.has_key('view'):
                    s.tree.expand_row(s.store.get_path(s.bind.views[z['view']]['node']),False)            
                s.tree.expand_row(zone_path,True)
                s.tree.set_cursor(zone_path)
                return True
            else:
                s.tree.collapse_row(zone_path)
                expanded=False
            for name in z['names'].keys():
                path=None                
                if type( z['names'][ name ] ) == list :
                    path = s.store.get_path(z['names'][ name ][0])
                    r = s.store.get_value(z['names'][ name ][0], RECORD_COLUMN)
                else:
                    path = s.store.get_path(z['names'][ name ])
                    r = s.store.get_value(z['names'][ name ], RECORD_COLUMN)
                rn=''
                if ( issubclass(r.__class__,RR) ):
                    if ( z['zone'].type == DNS.FORWARD ):                
                        rn = str(r.Name())
                    else:
                        rn = r.Name().label()
                if ( (len(name)>=lsv) and (name[:lsv] == sv)) or (
                     (len(rn)>=lsv) and (rn[:lsv] == sv)
                   ):
                    match=True                    
                    if not expanded:
                        debug( 'Expand Zone:',z['zone'].zone_origin, zone_path )
                        if z.has_key('view'):
                            s.tree.expand_row(s.store.get_path(s.bind.views[z['view']]['node']),False)
                        s.tree.expand_row(zone_path,False)
                        expanded = True
                        changed = True
                    debug( 'Expand Row:', name, path )
                    s.tree.expand_row(path,True)
                    s.tree.set_cursor(path)
                    changed = True
                elif s.tree.row_expanded(path):
                    s.tree.collapse_row(path)                    
        if not match and expanded:
            debug( 'Collapse Row:',  z['zone'].zone_origin, zone_path )
            s.tree.collapse_row(zone_path)
        
    def search_value_changed(s,w):
        sv=s.search_value.get_text()
        debug( 'SEARCH VALUE CHANGED:',sv )
        lsv=len(sv)
        if not s.bind.isViewed:
            for origin in s.bind.zones.keys():
                z = s.bind.zones[origin]
                s.set_zone_search_expansion( z, sv, lsv )
        else:
            for vn in s.bind.views.keys():
                v = s.bind.views[vn]
                s.tree.collapse_row(s.store.get_path(v['node']))
                for origin in v['zones'].keys():
                    z = v['zones'][origin]
                    s.set_zone_search_expansion(z,sv,lsv)
        return False
        
    def search_widget(s):
        w=gtk.HBox()
        l=gtk.Label(_('Search:'))
        w.pack_start(l,False)
        s.search_value = gtk.Entry()
        s.search_value.connect("changed",s.search_value_changed)
        w.pack_start(s.search_value,False)
        return w

    def toggle_show_dnssec(s,w):
        s.load()
        
    def dnssec_widget(s):
        s.showDNSSEC = gtk.CheckButton(label=_('Show DNSSEC Records'))
        s.showDNSSEC.set_active(False)
        s.showDNSSEC.connect("toggled", s.toggle_show_dnssec)
        return s.showDNSSEC
    
    def __init__(s, bind):
        s.bind = bind
        s.gui = bind.gui
        s.tree  = bind.gui.xml.get_widget('ZoneTree')
#        s.tree.set_headers_visible(False)
        s.tree.set_reorderable(True)
        s.tree.connect('drag-begin',s.drag_begin)
        s.tree.connect('drag-end',s.drag_end)        
        s.tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK,
                                        [('ZONE',gtk.TARGET_SAME_WIDGET,1)],
                                        gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_DEFAULT
                                       )
        s.tree.enable_model_drag_dest([('VIEW',gtk.TARGET_SAME_WIDGET,2)],
                                        gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_DEFAULT
                                     )
        # Can't get DnD to work using standard methods. The above just serves to disable
        # "standard" DnD - we'll do it our way. 
        bind.gui.xml.signal_autoconnect({'on_NewButton_clicked': s.NewClicked,
                                         'on_EditButton_clicked': s.EditClicked,
                                         'on_DeleteButton_clicked': s.DeleteClicked,
                                         'on_ImportButton_clicked': s.OpenClicked,
                                         'on_PreviewButton_clicked': s.PreviewClicked,
                                         'on_SaveButton_clicked': s.SaveClicked,
                                         'on_menu_save': s.SaveClicked,
                                         'on_menu_new': s.NewClicked,
                                         'on_menu_edit': s.EditClicked
                                        })
        s.newButton =  bind.gui.xml.get_widget('NewButton')
        newTT = gtk.Tooltips()
        s.newButton.set_tooltip(newTT,"New Item")
        newTT.enable()
        s.editButton =  bind.gui.xml.get_widget('EditButton')
        editTT= gtk.Tooltips()
        s.editButton.set_tooltip(editTT,"Edit Item")
        editTT.enable()
        s.deleteButton =  bind.gui.xml.get_widget('DeleteButton')
        deleteTT= gtk.Tooltips()
        s.deleteButton.set_tooltip(deleteTT,"Delete Item")
        deleteTT.enable()
        s.OpenButton =  bind.gui.xml.get_widget('OpenButton')
        OpenTT= gtk.Tooltips()
        s.OpenButton.set_tooltip(OpenTT,"Import Hosts File")
        OpenTT.enable()
        s.PreviewButton =  bind.gui.xml.get_widget('PreviewButton')
        PreviewTT= gtk.Tooltips()
        s.PreviewButton.set_tooltip(PreviewTT,"Preview Files")
        PreviewTT.enable()
        s.SaveButton =  bind.gui.xml.get_widget('SaveButton')
        SaveTT= gtk.Tooltips()
        s.SaveButton.set_tooltip(SaveTT,"Save Files")
        SaveTT.enable()
        s.appBar = bind.gui.xml.get_widget('AppBar')
        s.statusFrame=None
        for c in s.appBar.get_children():
            if c.__class__ == gtk.Frame:
                s.statusFrame = c
            debug( c)
        if( s.statusFrame != None):
            for c in s.statusFrame.get_children():
                s.statusFrame.remove(c)
            s.statusLabel = gtk.Label()
            s.statusFrame.add(s.statusLabel)
            s.statusFrame.show_all()
#        s.appBar.modify_text(gtk.gdk.color_parse("black"),"Status Message")
        s.statusLabel.set_markup('<small>'+_('Select an object; press and release the right-hand mouse button.')+'</small>')
        s.store = gtk.TreeStore(gtk.gdk.Pixbuf, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
        s.pixCell = gtk.CellRendererPixbuf()
        s.txtCell = gtk.CellRendererText()
        s.infoCell = gtk.CellRendererText()
        s.txtCell.set_property("markup",TEXT_COLUMN)
        s.infoCell.set_property("markup",TEXT_COLUMN)
        s.infoCell.set_property("size-points", 8)
        s.pixColumn = gtk.TreeViewColumn(None, s.pixCell, pixbuf=PIX_COLUMN)
        s.txtColumn = gtk.TreeViewColumn(None, s.txtCell, markup=TEXT_COLUMN)
        s.searchWidget=s.search_widget()
        s.last_path=(0,)
        s.last_expansion=False        
        s.txtColumn.set_widget( s.searchWidget )
        s.infoColumn= gtk.TreeViewColumn(None, s.infoCell, markup=INFO_COLUMN)
        s.showDNSSECWidget=s.dnssec_widget()
        s.infoColumn.set_widget( s.showDNSSECWidget )
        s.tree.append_column(s.pixColumn)
        s.tree.append_column(s.txtColumn)
        s.tree.append_column(s.infoColumn)
        s.tree.set_model(s.store)
        s.globe_pic=gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/world.png")
        s.view_pic=gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/view.png")
        s.master_zone_pic=gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/zone.png")
        s.zone_dnssec_pic=gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/zone_dnssec.png")
        s.slave_zone_pic=gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/slave.png")
        s.host_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/address.png")
        s.host6_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/address6.png")
        s.hostOSI_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/address_osi.png")
        s.reverse_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse.png")
        s.reverse_zone_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone.png")
        s.reverse_zone_dnssec_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone_dnssec.png")
        s.reverse_ip6_zone_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone_v6.png")
        s.reverse_ip6_zone_dnssec_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone_v6_dnssec.png")
        s.reverse_osi_zone_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone_osi.png")
        s.reverse_osi_zone_dnssec_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/reverse_zone_osi_dnssec.png")
        s.cname_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/cname.png")
        s.mx_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/mail.png")
        s.key_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/key.png")
        s.acl_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/filter.png")
        s.lock_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/lock.png")
        s.forward_zone_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/forward-zone.png")
        s.server_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/server.png")
        s.log_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/log.png")
        s.mail_pic = gtk.gdk.pixbuf_new_from_file(SCB + "/pixmaps/mail.png")
        s.tree.connect( 'row_activated', s.on_row_activated )
        s.tree.connect( 'cursor_changed', s.on_cursor_changed )
        s.tree.connect( 'button_release_event', s.popup_menu)
        s.searchWidget.show_all()
        s.load()
    
    def load(s):
        s.store.clear()
        s.zoneNodes={}
        s.n_views=0
        svr=s.store.append(None,[s.globe_pic, _('DNS Server'),'', None])
        if s.bind.nmdc.__contains__('controls'):
            try:
                s.store.append(svr,[s.server_pic, _('Server Controls'), '', DNS.Controls(s.bind.nmdc)])
            except DNS.ACL.Undefined, e:
                s.bind.gui.error(_('DNS Configuration Initialisation Error:'),
                                 _('named.conf "controls" statement contains a reference to an undefined ACL "%s".') % e.name,True
                                )
        if s.bind.nmdc.__contains__('logging'):
            s.store.append(svr,[s.log_pic,_('Logging Configuration'),'',DNS.Logging(s.bind.nmdc)])
        if s.bind.nmdc.__contains__('trusted-keys'):
            s.store.append(svr,[s.lock_pic,_('DNSSEC Trusted Keys'),'', DNSsec.TrustedKeys(s.bind)])
        for acl in s.bind.acls.keys():
            s.store.append(None,[s.acl_pic, acl, _('Access Control List Filter'), s.bind.acls[acl] ])
        for key in s.bind.keys.keys():
            s.store.append(None,[s.key_pic, key, RDNC_KEY_NAME, s.bind.keys[key] ])
        for svr in s.bind.servers.keys():
            s.store.append(None,[s.server_pic, svr, _('DNS Server') + ' ' + svr, s.bind.servers[svr]])
        if not s.bind.isViewed :                            
            for origin in s.bind.zones.keys():
                zone = s.bind.zones[origin]
                s.add_zone(origin, zone)
        else:
            for key in s.bind.views.keys():
                s.n_views += 1
                s.bind.views[key]['node'] = s.store.append(None,[s.view_pic, key, _('DNS View'), s.bind.views[key]['view'] ])
                view =  s.bind.views[key]['view']
                s.store.append( s.bind.views[key]['node'], [s.acl_pic, _('From: Filter'), _('Access Control List Filter'), view.inputFilter ] )
                s.store.append( s.bind.views[key]['node'], [s.acl_pic, _('To: Filter'), _('Access Control List Filter'), view.outputFilter ] )                
                for k in s.bind.views[key]['keys'].keys():
                    s.store.append(s.bind.views[key]['node'],[s.key_pic, k, RDNC_KEY_NAME, s.bind.views[key]['keys'][k] ])
                for svr in s.bind.views[key]['servers'].keys():
                    s.store.append(s.bind.views[key]['node'],[s.server_pic, svr, _('DNS Server') + ' ' + svr, s.bind.views[key]['servers'][svr]])
            for key in s.bind.views.keys():
                nz=0
                for origin in s.bind.views[key]['zones'].keys():
                    zone = s.bind.views[key]['zones'][origin]
                    s.add_zone(origin, zone)
                    nz+=1
                if nz > 0:
                    s.tree.expand_row(s.store.get_path(s.bind.views[key]['node']),False)
        if len(s.search_value.get_text()):
            s.search_value_changed(s.search_value)
        elif (s.last_path != (0,)):
            path=()
            depth=len(s.last_path)-1
            d=0
            for p in s.last_path:
                path+=(p,)
                try:
                    iter = s.store.get_iter(path)
                except ValueError:                    
                    p=path[-1]
                    path=path[:-1]
                    p-=1
                    path+=(p,)
                    try:
                        iter = s.store.get_iter(path)                        
                    except ValueError:
                        path = path[:-1]
                        break
                if( d < depth) or s.last_expansion:
                    s.tree.expand_row(path,False)
                d += 1
            s.tree.set_cursor(path)
        if hasattr(s.bind,'containsSignedZones') and s.bind.containsSignedZones:
            s.showDNSSECWidget.show_all()
            
    def add_zone(s, origin, z):
        zpic = s.master_zone_pic
        ztyp = ''
        if z['zone'].clas== 'IN':
            ztyp = _('Internet')
        elif z['zone'].clas=='HS':
            ztyp = 'Hesiod'
        elif z['zone'].clas=='CH':
            ztyp = 'CHAOS'
        elif z['zone'].clas=='CS':
            ztyp = 'CSNET'
        zoneSecure = False
        zoneContainsDNSSECRecords=False
        if z['zone'].cfgType == 'master':
            names={}
            has_key=False
            for r in  z['zone'].records:
                n = str(r.Name())
                t = r.Type()
                if not names.has_key(n):
                    names[n]={}
                elif t in DNSSEC_RRs:
                    names[n][t]=1
                    zoneContainsDNSSECRecords=True
                if t in ['KEY', 'DNSKEY']:
                    has_key=True
            if has_key:
                zoneSecure = True;
                for n in names.keys():
                    if not( ( names[n].has_key('RRSIG') and names[n].has_key('NSEC') ) or
                            ( names[n].has_key('SIG') and (names[n].has_key('NXT')))
                          ):
                        zoneSecure = False
        if zoneSecure:
            s.bind.containsSignedZones = True
            z['dnssecSecure']=True
        elif zoneContainsDNSSECRecords:
            z['dnssecSecure']=False
        if z['zone'].type == DNS.FORWARD:
            if not zoneSecure:
                ztyp +=  ' '+_('Forward')
            else:
                zpic = s.zone_dnssec_pic
                ztyp +=  ' '+_('Forward') + ' ' + _('(Signed)')
        elif z['zone'].type == DNS.IPV4.REVERSE:
            if not zoneSecure:
                zpic = s.reverse_zone_pic
                ztyp += ' '+_('Reverse') + ' ' + _('IPv4') 
            else:
                zpic = s.reverse_zone_dnssec_pic
                ztyp += ' '+_('Reverse') + ' ' + _('IPv4') + ' ' + _('(Signed)')            
        elif (z['zone'].type == DNS.IPV6.REVERSE_ARPA):
            if not zoneSecure:
                zpic = s.reverse_ip6_zone_pic
                ztyp += ' '+_('Reverse') + ' ' + _('IPv6 ARPA')
            else:
                zpic = s.reverse_ip6_zone_dnssec_pic
                ztyp += ' '+_('Reverse') + ' ' + _('IPv6 ARPA') + ' ' + _('(Signed)')                
        elif (z['zone'].type == DNS.IPV6.REVERSE_INT):
            if not zoneSecure:
                zpic = s.reverse_ip6_zone_pic                
                ztyp += ' '+ _('Reverse') + ' ' + _('IPv6 INT')
            else:
                zpic = s.reverse_ip6_zone_dnssec_pic
                ztyp += ' '+ _('Reverse') + ' ' + _('IPv6 INT') + ' ' + _('(Signed)')
        elif (z['zone'].type == DNS.NSAP.REVERSE):
            if not zoneSecure:
                zpic = s.reverse_osi_zone_pic                            
                ztyp += ' '+ _('Reverse') + ' ' + _('OSI NSAP')
            else:
                zpic = s.reverse_osi_zone_dnssec_pic                            
                ztyp += ' '+ _('Reverse') + ' ' + _('OSI NSAP') + ' ' + _('(Signed)')            
        if  z['config']['type'] in [ 'slave', 'stub' ]:
            zpic = s.slave_zone_pic
            ztyp += ' ' + _('Slave')
            if z['config']['type'] == 'stub':
                ztyp += ' ' + _('Stub')
        elif z['config']['type'] == 'forward':
            zpic = s.forward_zone_pic
            ztyp +=' ' + _('Forwarder')
        ztyp += ' ' + _('Zone')
        view = None
        if z.has_key('view'):
            if s.bind.views.has_key(z['view']) and s.bind.views[z['view']].has_key('node'):
                view = s.bind.views[z['view']]['node']
        z['node']=s.store.append(view, [ zpic, z['zone'].origin_name.label() , ztyp, ZoneNode(z) ])
#        debug( z['node'], z['zone'].zone_origin )
        z['names']={}
        for r in z['zone'].records:
            if not r.has_data() or ((not s.showDNSSEC.get_active()) and (r.Type() in DNSSEC_RRs)):
                continue
            s.add_record(z,r)

    def add_record(s, z, r):
        rtype = r.Type()
        if r.is_directive():
        #   Directives - should display something here
            return
        rname = r.Name().relative_label()
        node = z['node']
        rpic = s.host_pic
        header=''
        data=''
        if r.is_mapping():               
            rpic = s.host_pic
            if (rtype == 'AAAA') or (rtype == 'A6'):
                rpic = s.host6_pic
            elif rtype == 'CNAME':
                rpic = s.cname_pic
            elif (rtype == 'PTR'):
                rpic = s.host_pic
                if(z['zone'].type == DNS.IPV6.REVERSE_ARPA) or (z['zone'].type == DNS.IPV6.REVERSE_INT):
                    rpic=s.host6_pic
                elif( z['zone'].type == DNS.NSAP.REVERSE ):
                    rpic=s.hostOSI_pic
            mapping = r.attrs[0][0]
            t=type( r[mapping] )
            v=''
            if (t == int) or (t == str):
                v = r[mapping]
            else:
                v = r[mapping].label()
            if len(rname):
                header = rname + '\t->\t' + v
            else:  # (Top Level Domain!)      
                header = r.Name().label() + '\t->\t' + v
            data="<b>"+str(r.Name())+"</b>\n"
        else:
            rpic = s.reverse_pic
            if rtype == 'MX':
                rpic = s.mail_pic
            if rtype in DNSSEC_RRs:
                rpic = s.lock_pic
            header=rname + "\t" + r.desc['']
            if( r.Type() == 'NS' ):
                header += "\t" + r[r.attrs[0][0]].label()
            data="<b>"+str(r.Name())+"</b>\n"
        jlen=r.max_value_label() + 2
        jfmt='%-'+str(jlen)+'s'
	max_record_lenght = 30
        for name, val, wrap in r.value_labels():
            value_str = ''
            if wrap:
            	input_str = str(val)
	        start = 0
            	while start < len(input_str):
                  	if start == 0:
                  		data+=(jfmt % name)+": "+input_str[start:start+max_record_lenght]+'\n'
                  	else:
                  		data+=string.expandtabs('\t',jlen+2)+input_str[start:start+max_record_lenght]+'\n'
                  	start = start + max_record_lenght
	    else:
            	data+=(jfmt % name)+": "+str(val)+"\n"
        node=''
        if len( rname ) and ( rname != z['zone'].zone_origin ):            
            if z['names'].has_key(rname):
                node = s.store.append(z['names'][rname][0],[ rpic, header, r.Type(), r] )
                z['names'][rname].append(node)
            else:
                node=s.store.append(z['node'],[ rpic, header, r.Type(), r ] )
                z['names'][rname] = [node]
        else:
            node = s.store.append(z['node'],[ rpic, header, r.Type(), r ])
        s.store.append(node,[ s.reverse_pic, '<small><tt>'+data+'</tt></small>','', r] )
                                          
